#include "moo/macros.h"
#include "moo/appwindow.h"
#include "moo/app.h"
#include "moo/prefs.h"
#include "moo/windowmenu.h"
#include "moo/utils.h"
#include <QtGui>

using namespace moo;

static QList<AppWindow*> window_list;

struct moo::AppWindowPrivate {
    WindowMenu *windowMenu;
    bool initUiCalled;
    bool loadUiConfigCalled;

    AppWindowPrivate() :
        windowMenu(0),
        initUiCalled(false),
        loadUiConfigCalled(false)
    {
    }
};

QList<AppWindow*> AppWindow::listWindows()
{
    return window_list;
}

AppWindow::AppWindow(const QString &name) :
    impl(new AppWindowPrivate)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setAttribute(Qt::WA_QuitOnClose);
    setObjectName(name);

    WindowMenu::queueUpdateMenus();
    impl->windowMenu = new WindowMenu(this, this);
    window_list << this;
}

AppWindow::~AppWindow()
{
    window_list.removeAll(this);
    WindowMenu::queueUpdateMenus();
    delete impl;
}

WindowMenu *AppWindow::windowMenu()
{
    return impl->windowMenu;
}

void AppWindow::allApplyPrefs()
{
    foreach (AppWindow *w, window_list)
        w->applyPrefs();
}

void AppWindow::applyPrefs()
{
}

void AppWindow::showMessage(const QString &msg)
{
    QStatusBar *sb = findChild<QStatusBar*>("statusbar");
    m_return_if_fail(sb != 0);
    sb->showMessage(msg);
}

void AppWindow::clearMessage()
{
    QStatusBar *sb = findChild<QStatusBar*>("statusbar");
    m_return_if_fail(sb != 0);
    sb->clearMessage();
}

static QString windowSizeKey(AppWindow *w)
{
    return QString("ui/window-size-") + w->objectName();
}

static QString windowStateKey(AppWindow *w)
{
    return QString("ui/window-state-") + w->objectName();
}


static QAction *findAction(AppWindow *self, const char *name)
{
    QAction *a = self->findChild<QAction*>(name);
    if (!a)
        qCritical("no action `%s' in window `%s'", name,
                  qPrintable(self->objectName()));
    return a;
}

void AppWindow::connectAction(const char *name, const char *slot)
{
    if (QAction *a = findAction(this, name))
        connect(a, SIGNAL(triggered()), slot);
}

void AppWindow::connectAction(QAction *action, const char *slot)
{
    connect(action, SIGNAL(triggered()), slot);
}

void AppWindow::connectAction(const char *name, QObject *receiver, const char *slot)
{
    if (QAction *a = findAction(this, name))
        connect(a, SIGNAL(triggered()), receiver, slot);
}

void AppWindow::connectAction(QAction *action, QObject *receiver, const char *slot)
{
    connect(action, SIGNAL(triggered()), receiver, slot);
}

void AppWindow::setActionShortcuts(const char *action, const QList<QKeySequence> &shortcuts)
{
    if (QAction *a = findAction(this, action))
        a->setShortcuts(shortcuts);
}

void AppWindow::setActionShortcuts(const char *action, QKeySequence::StandardKey key)
{
    if (QAction *a = findAction(this, action))
        a->setShortcuts(key);
}

void AppWindow::initUi()
{
    if (impl->initUiCalled)
    {
        qCritical("%s called twice", Q_FUNC_INFO);
        return;
    }

    impl->initUiCalled = true;

    connectAction("actionPrint", SLOT(slotPrint()));
    connectAction("actionPageSetup", SLOT(slotPageSetup()));
    connectAction("actionPrintPreview", SLOT(slotPrintPreview()));

#if defined(Q_OS_MAC) || QT_VERSION < 0x040400
    if (QAction *a = findAction(this, "actionPrintPreview"))
        a->setVisible(false);
#endif

    connectAction("actionClose", SLOT(close()));

    connectAction("actionAbout", mooApp, SLOT(aboutDialog()));
    connectAction("actionAboutQt", mooApp, SLOT(aboutQt()));
    connectAction("actionPrefs", mooApp, SLOT(prefsDialog()));
    connectAction("actionQuit", mooApp, SLOT(maybeQuit()));

    setActionShortcuts("actionClose", QKeySequence::Close);

    loadUiConfig();
}

void AppWindow::loadUiConfig()
{
    if (impl->loadUiConfigCalled)
    {
        qCritical("%s called twice", Q_FUNC_INFO);
        return;
    }

    impl->loadUiConfigCalled = true;

    QSize size = stateValue(windowSizeKey(this), QSize()).toSize();
    if (size.width() > 0 && size.height() > 0)
        resize(size);

    QByteArray state = stateValue(windowStateKey(this), QByteArray()).toByteArray();
    if (!state.isEmpty())
        restoreState(state);
}

void AppWindow::saveUiConfig()
{
    setStateValue(windowSizeKey(this), size());
    setStateValue(windowStateKey(this), saveState());
}

void AppWindow::showEvent(QShowEvent *e)
{
    if (!impl->initUiCalled)
        qCritical("AppWindow::initUi() has not been called");
    QMainWindow::showEvent(e);
}

void AppWindow::closeEvent(QCloseEvent *e)
{
    saveUiConfig();
    e->accept();
}

void AppWindow::zoomWindow()
{
    if (isMaximized())
        showNormal();
    else
        showMaximized();
}

void AppWindow::setWindowTitle(const QString &title)
{
    QMainWindow::setWindowTitle(title);
    emit windowTitleChanged(title);
}

void AppWindow::setWindowFileUrl(const QUrl &url)
{
#if QT_VERSION >= 0x040400
    QString filename = url.toLocalFile();
    if (filename.isEmpty())
        filename = url.toString();
    QMainWindow::setWindowFilePath(filename);
    emit windowFileUrlChanged(url);
#else
    setWindowTitle(util::displayName(url));
#endif
}

void AppWindow::setWindowModified(bool modified)
{
    QMainWindow::setWindowModified(modified);
    emit windowModifiedChanged(modified);
}


///////////////////////////////////////////////////////////////////////////
//
// Printing
//

namespace {

class PageSetup {
    bool isValid;
    QPrinter::Orientation orientation;
    QPrinter::PageSize size;

public:
    PageSetup(QPrinter *printer) :
        isValid(false),
        orientation(QPrinter::Portrait),
        size(QPrinter::A4)
    {
        m_return_if_fail(printer != 0);
        orientation = printer->orientation();
        size = printer->pageSize();
        isValid = true;
    }

    void apply(QPrinter *printer)
    {
        m_return_if_fail(printer != 0);
        if (isValid)
        {
            printer->setPageSize(size);
            printer->setOrientation(orientation);
        }
    }

    PageSetup(const QByteArray &data) :
        isValid(false),
        orientation(QPrinter::Portrait),
        size(QPrinter::A4)
    {
        QDataStream s(data);
        QMap<QString, QVariant> map;
        s >> map;
        QString version = map["version"].toString();
        if (version != "1.0")
        {
            qCritical("%s: oops", Q_FUNC_INFO);
            return;
        }
        orientation = QPrinter::Orientation(map["orientation"].toInt());
        size = QPrinter::PageSize(map["size"].toInt());
        isValid = true;
    }

    QByteArray save() const
    {
        if (!isValid)
            return QByteArray();

        QMap<QString, QVariant> map;
        map["version"] = QString("1.0");
        map["orientation"] = int(orientation);
        map["size"] = int(size);
        QByteArray data;
        QDataStream ds(&data, QIODevice::WriteOnly);
        ds << map;
        return data;
    }
};

static QPrinter *createPrinter()
{
    QPrinter *printer = new QPrinter;
    QByteArray data = prefsValue(Prefs::PageSetup);

    if (!data.isEmpty())
    {
        PageSetup ps(data);
        ps.apply(printer);
    }

    return printer;
}

static moo::util::DataOnDemand<QPrinter> global_printer(createPrinter);

}

static void savePrinter()
{
    m_return_if_fail(global_printer);
    PageSetup ps(global_printer);
    setPrefsValue(Prefs::PageSetup, ps.save());
}

// void AppWindow::printDocument()
// {
//     QTextEdit *doc = textEdit();
//     m_return_if_fail(doc != 0);
//
//     QPrintDialog dlg(global_printer, this);
//     if (doc->textCursor().hasSelection())
//         dlg.addEnabledOption(QAbstractPrintDialog::PrintSelection);
//     if (dlg.exec() != QDialog::Accepted)
//         return;
//
//     doc->print(global_printer);
//
//     savePrinter();
// }

// void AppWindow::previewPaintRequested(QPrinter *printer)
// {
//     QTextEdit *doc = textEdit();
//     m_return_if_fail(doc != 0);
//     doc->print(printer);
// }

void AppWindow::printDocument(QPrinter *)
{
    m_return_if_reached();
}

void AppWindow::setupPrintDialog(QPrintDialog *)
{
    m_return_if_reached();
}

void AppWindow::slotPrint()
{
    QPrintDialog dlg(global_printer, this);
    setupPrintDialog(&dlg);

    if (dlg.exec() != QDialog::Accepted)
        return;

    printDocument(global_printer);
    savePrinter();
}

void AppWindow::slotPreviewPaintRequested(QPrinter *printer)
{
    printDocument(printer);
}

void AppWindow::slotPrintPreview()
{
#if QT_VERSION >= 0x040400
    QPrintPreviewDialog dlg(global_printer, this);
    connect(&dlg, SIGNAL(paintRequested(QPrinter*)),
            SLOT(slotPreviewPaintRequested(QPrinter*)));
    dlg.exec();
#endif
}

void AppWindow::slotPageSetup()
{
    QPageSetupDialog dlg(global_printer, this);
    if (dlg.exec() == QDialog::Accepted)
        savePrinter();
}
